import contextlib
import io
import os
import tarfile
import time
from pathlib import Path
from zipfile import ZipFile

import pytest
from conda_package_handling.validate import validate_converted_files_match_streaming

from conda_package_streaming.package_streaming import (
    CondaComponent,
    stream_conda_component,
)
from conda_package_streaming.transmute import transmute, transmute_tar_bz2


@pytest.fixture
def testtar_bytes():
    buffer = io.BytesIO()
    with tarfile.open("test.tar.bz2", "w:bz2", fileobj=buffer) as tar:
        symlink = tarfile.TarInfo(name="symlink")
        symlink.type = tarfile.LNKTYPE
        symlink.linkname = "target"
        tar.addfile(symlink)

        expected = tarfile.TarInfo(name="info/expected")
        tar.addfile(expected, io.BytesIO())
        unexpected = tarfile.TarInfo(name="info/unexpected")
        tar.addfile(unexpected, io.BytesIO())
    return buffer.getbuffer()


@contextlib.contextmanager
def timeme(message: str = ""):
    begin = time.time()
    yield
    end = time.time()
    print(f"{message}{end-begin:0.2f}s")


def test_transmute(conda_paths: list[Path], tmpdir):
    tarbz_packages = []
    for path in conda_paths:
        path = str(path)
        if path.endswith(".tar.bz2") and (1 << 20 < os.stat(path).st_size < 1 << 22):
            tarbz_packages = [path]
    conda_packages = []  # not supported

    assert tarbz_packages, "no medium-sized .tar.bz2 packages found"

    metadata_checks = 0

    for packages in (conda_packages, tarbz_packages):
        for package in packages:
            with timeme(f"{package} took "):
                out = transmute(package, tmpdir)
                _, missing, mismatched = validate_converted_files_match_streaming(
                    out, package, strict=True
                )
                assert missing == mismatched == []
                if out.name.endswith(".conda"):
                    with ZipFile(out) as zf:
                        metadata_checks += 1
                        assert "metadata.json" in zf.namelist()

    assert metadata_checks > 0


def test_transmute_symlink(tmpdir, testtar_bytes):
    testtar = Path(tmpdir, "test.tar.bz2")
    testtar.write_bytes(testtar_bytes)

    out = transmute(str(testtar), tmpdir)
    _, missing, mismatched = validate_converted_files_match_streaming(
        out, testtar, strict=True
    )
    assert missing == mismatched == []


def test_transmute_info_filter(tmpdir, testtar_bytes):
    testtar = Path(tmpdir, "test.tar.bz2")
    testtar.write_bytes(testtar_bytes)

    transmute(
        str(testtar), tmpdir, is_info=lambda filename: filename == "info/expected"
    )

    with open(Path(tmpdir, "test.conda"), "rb") as fileobj:
        for component, expected in (CondaComponent.info, {"info/expected"}), (
            CondaComponent.pkg,
            {
                "info/unexpected",
                "symlink",
            },
        ):
            items = stream_conda_component("test.conda", fileobj, component)
            assert {member.name for tar, member in items} == expected, items


def test_transmute_backwards(tmpdir, conda_paths):
    tarbz_packages = []
    for path in conda_paths:
        path = str(path)
        if path.endswith(".conda") and (1 << 20 < os.stat(path).st_size < 1 << 22):
            tarbz_packages = [path]
    conda_packages = []  # not supported

    assert tarbz_packages, "no medium-sized .conda packages found"

    for packages in (conda_packages, tarbz_packages):
        for package in packages:
            with timeme(f"{package} took "):
                out = transmute_tar_bz2(package, tmpdir)
                _, missing, mismatched = validate_converted_files_match_streaming(
                    out, package, strict=True
                )
                assert missing == mismatched == []


def test_transmute_tarbz2_to_tarbz2(tmpdir, testtar_bytes):
    testtar = Path(tmpdir, "test.tar.bz2")
    testtar.write_bytes(testtar_bytes)
    outdir = Path(tmpdir, "output")
    outdir.mkdir()
    out = transmute_tar_bz2(str(testtar), outdir)
    _, missing, mismatched = validate_converted_files_match_streaming(
        out, testtar, strict=True
    )
    assert missing == mismatched == []



import boto3
import pytest

from conda_package_streaming import s3

LIMIT = 16


@pytest.fixture
def s3_client(package_server):
    host, port = package_server.server.server_address
    client = boto3.client(
        "s3",
        aws_access_key_id="test_id",
        aws_secret_access_key="test_key",
        endpoint_url=f"http://{host}:{port}",
        use_ssl=False,
        verify=False,
    )
    return client


def test_head_objects(s3_client, conda_paths):
    bucket = "pkgs"  # independent of filesystem path
    for path in conda_paths[:LIMIT]:
        s3_client.head_object(Bucket=bucket, Key=path.name)


def test_stream_s3(s3_client, conda_paths):
    with pytest.raises(ValueError):
        next(s3.stream_conda_info(s3_client, "pkgs", "notaconda.rar"))

    for path in conda_paths[:LIMIT]:
        members = s3.stream_conda_info(s3_client, "pkgs", path.name)
        print("stream s3", path.name)
        for tar, member in members:
            if member.name == "info/index.json":
                members.close()  # faster than waiting for gc?
                break
        else:
            pytest.fail("info/index.json not found")



"""
Test web server.
"""
import logging
import threading
import wsgiref.simple_server
from pathlib import Path
from typing import Any

import bottle
import conftest

log = logging.getLogger(__name__)


def get_app(pkgs_dir):
    """
    Bottle conveniently supports Range requests.

    Server may block if browser etc. keeps connection open.
    """
    app = bottle.Bottle()
    app.pkgs_dir = pkgs_dir

    def serve_file(filename):
        mimetype = "auto"
        # from https://repo.anaconda.com/ behavior:
        if filename.endswith(".tar.bz2"):
            mimetype = "application/x-tar"
        elif filename.endswith(".conda"):
            mimetype = "binary/octet-stream"
        return bottle.static_file(filename, root=pkgs_dir, mimetype=mimetype)

    app.route("/pkgs/<filename>", "GET", serve_file)
    return app


def selftest():
    """
    Run server in a thread that will die when the application exits.
    """
    t = get_server_thread(conftest.find_packages_dirs())
    t.start()

    import time

    time.sleep(300)


class ServerThread(threading.Thread):
    server: wsgiref.simple_server.WSGIServer
    app: Any


def get_server_thread(pkgs_dir: Path):
    """
    Return test server thread with additional .server, .app properties.

    Call .start() to serve in the background.
    """
    app = get_app(pkgs_dir)
    server = wsgiref.simple_server.make_server("127.0.0.1", 0, app)
    log.info(f"serving {app.pkgs_dir} on {server.server_address}/pkgs")
    t = ServerThread(daemon=True, target=server.serve_forever)
    t.app = app
    t.server = server  # server.application == app
    return t


if __name__ == "__main__":
    import logging

    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)s %(message)s",
        datefmt="%Y-%m-%dT%H:%M:%S",
    )
    log.setLevel(logging.DEBUG)
    selftest()



import io
import stat
import tarfile
from errno import ELOOP

import pytest

from conda_package_streaming import exceptions, extract, package_streaming

MAX_CONDAS = 8


def test_extract_stream(conda_paths, tmp_path):
    for i, package in enumerate(conda_paths):
        print(package)
        with open(package, "rb") as fileobj:
            stream = package_streaming.stream_conda_component(
                package, fileobj, component=package_streaming.CondaComponent.pkg
            )
            dest_dir = tmp_path / package.name
            extract.extract_stream(stream, dest_dir)

        if i >= MAX_CONDAS:
            break


def test_extract_all(conda_paths, tmp_path):
    for i, package in enumerate(conda_paths):
        print(package)
        dest_dir = tmp_path / package.name
        extract.extract(package, dest_dir=dest_dir)

        if i >= MAX_CONDAS:
            break


def empty_tarfile(name, mode=0o644):
    """
    Return BytesIO containing a tarfile with one empty file named :name
    """
    tar = io.BytesIO()
    t = tarfile.TarFile(mode="w", fileobj=tar)
    tarinfo = tarfile.TarInfo(name=name)
    tarinfo.mode = mode
    t.addfile(tarinfo, io.BytesIO())
    t.close()
    tar.seek(0)
    return tar


def test_oserror(tmp_path):
    """
    Fail if tarfile raises OSError (formerly known as IOError)
    """
    tar = empty_tarfile("empty-test")

    class TarELOOP(tarfile.TarFile):
        def extractall(self, path=None, members=None):
            raise OSError(ELOOP, "case sensitivity")

    class TarOSError(tarfile.TarFile):
        def extractall(self, path=None, members=None):
            raise OSError("not eloop")

    def stream(cls):
        yield (cls(fileobj=tar), tarfile.TarInfo())

    with pytest.raises(exceptions.CaseInsensitiveFileSystemError):
        extract.extract_stream(stream(TarELOOP), tmp_path)

    with pytest.raises(OSError):
        extract.extract_stream(stream(TarOSError), tmp_path)


def stream(fileobj):
    """
    Like the tuples produced by part of conda-package-streaming.
    """
    yield (package_streaming.TarfileNoSameOwner(fileobj=fileobj), tarfile.TarInfo())


def stream_stdlib(fileobj):
    """
    Like the tuples produced by part of conda-package-streaming.
    """
    yield (tarfile.TarFile(fileobj=fileobj), tarfile.TarInfo())


def test_slip(tmp_path):
    """
    Fail if tarfile tries to put files outside its dest_dir (tmp_path)
    """

    tar = empty_tarfile(name="../slip")

    with pytest.raises(exceptions.SafetyError):
        extract.extract_stream(stream(tar), tmp_path)

    tar2 = empty_tarfile(name="/absolute")

    with pytest.raises(exceptions.SafetyError):
        extract.extract_stream(stream(tar2), tmp_path)


def test_chown(conda_paths, tmp_path, mocker):
    for package in conda_paths[:2]:
        print(package)
        with open(package, "rb") as fileobj:
            stream = package_streaming.stream_conda_component(
                package, fileobj, component=package_streaming.CondaComponent.pkg
            )

            for tar, member in stream:
                assert isinstance(tar, package_streaming.TarfileNoSameOwner), tar
                break


def test_umask(tmp_path, mocker):
    """
    Demonstrate that umask-respecting tar implementation works.

    Mock umask in case it is different on your system.
    """
    mocker.patch("conda_package_streaming.package_streaming.UMASK", new=0o22)

    tar3 = empty_tarfile(name="naughty_umask", mode=0o777)
    extract.extract_stream(stream_stdlib(tar3), tmp_path)
    mode = (tmp_path / "naughty_umask").stat().st_mode
    assert mode & stat.S_IWGRP, "%o" % mode

    tar3.seek(0)
    extract.extract_stream(stream(tar3), tmp_path)
    mode = (tmp_path / "naughty_umask").stat().st_mode
    assert not mode & stat.S_IWGRP, "%o" % mode



import json

import pytest

from conda_package_streaming import package_streaming


def test_package_streaming(conda_paths):
    for path in conda_paths:
        if str(path).endswith(".conda"):
            with pytest.raises(LookupError):
                package_streaming.stream_conda_component(path, component="notfound")

        with pytest.raises(ValueError):
            package_streaming.stream_conda_component("notapackage.rar")


def test_early_exit(conda_paths):
    for package in conda_paths:
        print(package)
        stream = iter(package_streaming.stream_conda_info(package))
        found = False
        for tar, member in stream:
            assert not found, "early exit did not work"
            if member.name == "info/index.json":
                reader = tar.extractfile(member)
                if reader:
                    json.load(reader)
                    found = True
                stream.close()  # PEP 342 close()
        # stream_conda_info doesn't close a passed-in fileobj, but a
        # filename should be closed.
        assert found, f"index.json not found in {package}"



import io
import tempfile
from contextlib import closing, contextmanager
from pathlib import Path
from zipfile import ZipFile

import pytest
from requests import HTTPError, Session

from conda_package_streaming import lazy_wheel
from conda_package_streaming.lazy_wheel import LazyConda
from conda_package_streaming.url import (
    conda_reader_for_url,
    extract_conda_info,
    stream_conda_info,
)

LIMIT = 16


@pytest.fixture
def package_url(package_server):
    """
    Base url for all test packages.
    """
    host, port = package_server.server.server_address
    return f"http://{host}:{port}/pkgs"


@pytest.fixture
def package_urls(package_server, package_url):
    pkgs_dir = Path(package_server.app.pkgs_dir)
    conda = []
    tar_bz2 = []

    for path in pkgs_dir.iterdir():
        if len(conda) > LIMIT and len(tar_bz2) > LIMIT:
            break
        url = f"{package_url}/{path.name}"
        if path.name.endswith(".tar.bz2") and len(tar_bz2) < LIMIT:
            tar_bz2.append(url)
        elif path.name.endswith(".conda") and len(conda) < LIMIT:
            conda.append(url)
    # interleave
    urls = []
    for pair in zip(conda, tar_bz2):
        urls.extend(pair)
    return urls


def test_stream_url(package_urls):
    with pytest.raises(ValueError):
        next(stream_conda_info("https://localhost/notaconda.rar"))

    for url in package_urls:
        with closing(stream_conda_info(url)) as members:
            print("stream_url", url)
            for tar, member in members:
                if member.name == "info/index.json":
                    break
            else:
                pytest.fail("info/index.json not found")


def test_fetch_meta(package_urls):
    for url in package_urls:
        with tempfile.TemporaryDirectory() as destdir:
            extract_conda_info(url, destdir)


def test_lazy_wheel(package_urls):
    lazy_tests = 7
    for url in package_urls:
        if url.endswith(".conda"):
            # API works with `.tar.bz2` but only returns LazyConda for `.conda`
            filename, conda = conda_reader_for_url(url)
            assert filename == url.rsplit("/")[-1]
            with conda:
                assert isinstance(conda, LazyConda)
                assert conda.mode == "rb"
                assert conda.readable()
                assert not conda.writable()
                assert not conda.closed

                request_count = conda._request_count

                # did we really prefetch the info?
                zf = ZipFile(conda)  # type: ignore
                filename = filename[: -len(".conda")]
                zf.open(f"info-{filename}.tar.zst").read()

                assert (
                    conda._request_count == request_count
                ), "info required extra GET request"
                assert conda._request_count <= 3

                conda.prefetch("not-appearing-in-archive.txt")

                # zip will figure this out naturally; delete method?
                conda._check_zip()

            lazy_tests -= 1
            if lazy_tests <= 0:
                break
    else:
        raise LookupError(
            "not enough .conda packages found %d %s" % (lazy_tests, package_urls)
        )

    with pytest.raises(HTTPError):
        conda_reader_for_url(package_urls[0] + ".404.conda")

    class Session200(Session):
        def get(self, *args, **kwargs):
            response = super().get(*args, **kwargs)
            response.status_code = 200
            return response

    with pytest.raises(lazy_wheel.HTTPRangeRequestUnsupported):
        LazyConda(package_urls[0], Session200())

    for url in package_urls:
        if url.endswith(".tar.bz2"):
            LazyConda(url, Session())._check_zip()
            break
    else:
        raise LookupError("no .tar.bz2 packages found")


def test_no_file_after_info():
    """
    If info is the last file, LazyConda must fetch (start of info file .. start
    of zip directory) instead of to the next file in the zip.
    """

    class MockBytesIO(io.BytesIO):
        prefetch = LazyConda.prefetch

        @contextmanager
        def _stay(self):
            yield

    zip = MockBytesIO()
    zf = ZipFile(zip, "w")
    zf.writestr("info-test.tar.zst", b"00000000")  # a short file
    zf.close()

    zip.prefetch("test")


@pytest.mark.skip()
def test_obsolete_lazy_wheel_selftest():
    import logging

    import requests

    logging.basicConfig(level=logging.DEBUG)

    session = requests.Session()

    lzoh = lazy_wheel.LazyZipOverHTTP(
        "https://repodata.fly.dev/repo.anaconda.com/pkgs/main/win-32/current_repodata.jlap",
        session,
    )

    lzoh.seek(1024)
    lzoh.read(768)
    lzoh.seek(0)

    # compare against regular fetch
    with open("outfile.txt", "wb+") as out:
        buf = b" "
        while buf:
            buf = lzoh.read(1024 * 10)
            print(list(zip(lzoh._left, lzoh._right)), lzoh._length)
            if not buf:
                break
            out.write(buf)



import nox


@nox.session(venv_backend="conda")
@nox.parametrize(
    "python",
    [(python) for python in ("3.7", "3.8", "3.9", "3.10")],
)
def tests(session):
    session.install("-e", ".[test]")
    session.run("pytest")
